#include "../Framework2/OpenGL/OpenGLApp.h"
#include "../Framework2/Util/SpringSystem.h"
#include "../Framework2/Math/PerlinNoise.h"

#include "../Framework2/Util/OpenGLModel.h"

#if defined(LINUX)
#include <dirent.h>
#define min(x, y) ((x < y)? x : y)
#endif

#define CLOTH_SIZE_X 48
#define CLOTH_SIZE_Y 42
#define C_SIZE (320.0f / CLOTH_SIZE_X)

struct Vertex {
	vec3 pos;
	vec3 normal;
	vec2 texCoord;
};


class MainApp : public OpenGLApp {
public:
	void resetCamera();

	bool init();
	bool exit();

	bool load();
	bool unload();

	void drawLight();
	void drawPole();
	void drawArrow();
	void drawCloth();
	bool drawFrame();
protected:
	SpringSystem spring;
	vec3 wind, wPos;

	Vertex vertices[CLOTH_SIZE_Y][CLOTH_SIZE_X];
	unsigned short indices[CLOTH_SIZE_X * 2];
	vec3 lightPos;

	float angle, rad;

	float nextTime;
	int flagState;

	ShaderID lighting, lighting2;

	TextureID base, light;
	Set <TextureID> flags;
};

void MainApp::resetCamera(){
	position = vec3(0, 0, -490);
	wx = 0;
	wy = 0;
	wz = 0;
}

bool MainApp::init(){
	unsigned int i, j;

	nextTime = 0;
	srand((unsigned int) cpuHz);

	initPerlin();

	for (j = 0; j < CLOTH_SIZE_X; j++){
		indices[2 * j] = CLOTH_SIZE_X + j;
		indices[2 * j + 1] = j;
	}


	for (j = 0; j < CLOTH_SIZE_Y; j++){
		for (i = 0; i < CLOTH_SIZE_X; i++){
			vertices[j][i].pos = vec3(C_SIZE * i, C_SIZE * (CLOTH_SIZE_Y - 1 - j), 0);
			vertices[j][i].texCoord = vec2(float(i) / (CLOTH_SIZE_X - 1), float(j) / (CLOTH_SIZE_Y - 1));
		}
	}

	spring.addRectField(CLOTH_SIZE_X, CLOTH_SIZE_Y, vertices, ((char *) vertices) + sizeof(vec3), sizeof(Vertex));

	for (i = 0; i < CLOTH_SIZE_X * CLOTH_SIZE_Y; i += CLOTH_SIZE_X) spring.getNode(i)->locked = true;

	wPos = vec3(0,0,0);
	flagState = -1;
	return true;
}

bool MainApp::exit(){
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}

	if ((lighting  = renderer->addShader("lighting.shd" )) == SHADER_NONE) return false;
	if ((lighting2 = renderer->addShader("lighting2.shd")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((light = renderer->addTexture("../Textures/Particle.tga", TEX_CLAMP)) == TEXTURE_NONE) return false;

#if defined(_WIN32)

	int result;
	WIN32_FIND_DATA findFileData;
	HANDLE handle = FindFirstFile("../Textures/Flags/*.*", &findFileData);
	if (handle != INVALID_HANDLE_VALUE){
		do {
			if (findFileData.cFileName[0] != '.'){
				TextureID tex = renderer->addTexture(String("../Textures/Flags/") + findFileData.cFileName, TEX_CLAMP);
				if (tex != TEXTURE_NONE) flags.add(tex);
			}
			result = FindNextFile(handle, &findFileData);
		} while (result > 0);
		FindClose(handle);
	}

#elif defined(LINUX)

	DIR *dir = opendir("../Textures/Flags/");
	if (dir != NULL){
		dirent *entry;
		while ((entry = readdir(dir)) != NULL){
			if (entry->d_name[0] != '.'){
				TextureID tex = renderer->addTexture(String("../Textures/Flags/") + entry->d_name, TEX_CLAMP);
				if (tex != TEXTURE_NONE) flags.add(tex);
			}
		}
	}

#endif

	if (flags.getCount() == 0) return false;
	base = flags[rand() % flags.getCount()];

	return true;
}

bool MainApp::unload(){
	return true;
}

void MainApp::drawLight(){
	renderer->setBlending(GL_ONE, GL_ONE);
	renderer->setTextures(light);
	renderer->apply();

	glColor3f(1,1,1);

	vec3 dx(modelView.elem[0][0], modelView.elem[0][1], modelView.elem[0][2]);
	vec3 dy(modelView.elem[1][0], modelView.elem[1][1], modelView.elem[1][2]);
	glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3fv(lightPos - 15 * dx + 15 * dy);

		glTexCoord2f(1,0);
		glVertex3fv(lightPos + 15 * dx + 15 * dy);

		glTexCoord2f(1,1);
		glVertex3fv(lightPos + 15 * dx - 15 * dy);

		glTexCoord2f(0,1);
		glVertex3fv(lightPos - 15 * dx - 15 * dy);
	glEnd();
}

void MainApp::drawPole(){
	renderer->setShader(lighting2);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant3f("camPos", position);
	renderer->changeShaderConstant4f("color", vec4(0.9f, 0.9f, 0.9f, 0));

	glBegin(GL_TRIANGLE_STRIP);
	for (unsigned int i = 0; i <= 16; i++){
		vec3 dir(cosf(i * PI / 8), 0, sinf(i * PI / 8));

		glNormal3fv(dir);
		glVertex3fv(11 * dir + vec3(0, C_SIZE * (CLOTH_SIZE_Y - 1) + 10, 0));
		glVertex3fv(11 * dir + vec3(0, -750, 0));
	}
	glEnd();
}

void MainApp::drawArrow(){
	renderer->setShader(lighting2);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant3f("camPos", position);
	renderer->changeShaderConstant4f("color", vec4(1, 0, 1, 0));

    glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glRotatef(-angle * 180 / PI, 0, 1, 0);

	glBegin(GL_TRIANGLE_FAN);
		glNormal3f(0, 1, 0);

		glVertex3f(rad * 80 + 40, C_SIZE * (CLOTH_SIZE_Y - 1) + 5,  0);
		glVertex3f(rad * 80,      C_SIZE * (CLOTH_SIZE_Y - 1) + 5,  30);
		glVertex3f(rad * 80,      C_SIZE * (CLOTH_SIZE_Y - 1) + 5,  10);
		glVertex3f(0,             C_SIZE * (CLOTH_SIZE_Y - 1) + 5,  10);
		glVertex3f(0,             C_SIZE * (CLOTH_SIZE_Y - 1) + 5, -10);
		glVertex3f(rad * 80,      C_SIZE * (CLOTH_SIZE_Y - 1) + 5, -10);
		glVertex3f(rad * 80,      C_SIZE * (CLOTH_SIZE_Y - 1) + 5, -30);
	glEnd();

	glPopMatrix();
}

void MainApp::drawCloth(){
	renderer->setShader(lighting);
	renderer->setTexture("Base", base);
	renderer->apply();

	renderer->changeShaderConstant3f("lightPos", lightPos);
	renderer->changeShaderConstant3f("camPos", position);

	glClientActiveTextureARB(GL_TEXTURE0_ARB);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	for (unsigned int i = 0; i < CLOTH_SIZE_Y - 1; i++){
		char *base = ((char *) vertices) + i * CLOTH_SIZE_X * sizeof(Vertex);
		glVertexPointer  (3, GL_FLOAT, sizeof(Vertex), base);
		glNormalPointer  (   GL_FLOAT, sizeof(Vertex), base + sizeof(vec3));
		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), base + sizeof(vec3) * 2);

		glDrawElements(GL_TRIANGLE_STRIP, CLOTH_SIZE_X * 2, GL_UNSIGNED_SHORT, indices);
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

void processNode(SNode *node, float *attribs){
	vec3 pos = (*(vec3 *) (attribs + 1) + *node->pos) * 0.00432f;

	node->dir += 500 * attribs[0] * vec3(noise3(pos.x, pos.y, pos.z), noise3(pos.x + 5.571f, pos.y - 9.842f, pos.z + 4.741f), noise3(pos.x - 2.3142f, pos.y + 7.1423f, pos.z + 3.412f));
}

#define T_DIFF 9

bool MainApp::drawFrame(){
	renderer->changeMask(ALL);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	projection = projectionMatrixX(1.5f, float(height) / float(width), 0.1f, 6000);
	glLoadMatrixf(transpose(projection));

	modelView = rotateZXY(-wx, -wy, -wz);
	modelView.translate(-position);

    glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(transpose(modelView));

	lightPos = vec3(3 * C_SIZE * CLOTH_SIZE_X * noise1(0.5f * time), 320 + 100 * noise1(0.3f * time + 23.37f), 3 * C_SIZE * CLOTH_SIZE_X * noise1(12.31f - 0.5f * time));

	if (time >= nextTime){
		nextTime = time + T_DIFF;
		flagState++;

		if (flagState < 5){
			float off = 4 * (rand() * (1.0f / RAND_MAX) - 0.5f);
			angle += off + sign(off) * 0.5f;
			rad = rand() * (0.5f / RAND_MAX) + 0.3f;

			wind = vec3(rad * cosf(angle), 0, rad * sinf(angle));
			spring.setGravity(350 * wind - vec3(0, 70, 0));
		} else if (flagState == 6){
			base = flags[rand() % flags.getCount()];
		} else if (flagState == 7){
			flagState = 0;
		}
	}

	if (flagState == 5){
		float off = 750.0f * sCurve(sCurve((T_DIFF - nextTime + time) / T_DIFF));
		for (int i = 0; i < CLOTH_SIZE_Y; i++){
			spring.getNode(i * CLOTH_SIZE_X)->pos->y = C_SIZE * (CLOTH_SIZE_Y - 1 - i) - off;
		}
	} else if (flagState == 6){
		float off = 750.0f * sCurve(sCurve((nextTime - time) / T_DIFF));
		for (int i = 0; i < CLOTH_SIZE_Y; i++){
			spring.getNode(i * CLOTH_SIZE_X)->pos->y = C_SIZE * (CLOTH_SIZE_Y - 1 - i) - off;
		}
	}

	float dTime = min(frameTime, 0.01f);
	wPos -= 500 * dTime * wind;
	float attribs[] = { dTime * rad, wPos.x, wPos.y, wPos.z };
	spring.update(dTime, processNode, attribs);

	spring.evaluateNormals();


	drawCloth();

	drawPole();
	drawArrow();


	drawLight();

	return true;
}

App *app = new MainApp();
